(function(global, factory) {
  typeof exports === "object" && typeof module !== "undefined" ? factory(exports) : typeof define === "function" && define.amd ? define(["exports"], factory) : (global = typeof globalThis !== "undefined" ? globalThis : global || self, factory(global.utils = {}));
})(this, function(exports2) {
  "use strict";var __defProp = Object.defineProperty;
var __defNormalProp = (obj, key, value) => key in obj ? __defProp(obj, key, { enumerable: true, configurable: true, writable: true, value }) : obj[key] = value;
var __publicField = (obj, key, value) => __defNormalProp(obj, typeof key !== "symbol" ? key + "" : key, value);

  function clean(obj, undefinedOnly = false) {
    if (obj == null) throw new Error("Cannot clean a NULL value");
    if (Array.isArray(obj)) {
      obj = obj.filter((o) => o != null);
    } else {
      Object.entries(obj).forEach(([key, value]) => {
        if (undefinedOnly && value === void 0 || !undefinedOnly && value == null) delete obj[key];
      });
    }
    return obj;
  }
  function deepCopy(value) {
    try {
      return structuredClone(value);
    } catch {
      return JSON.parse(JSONSanitize(value));
    }
  }
  function deepMerge(target, ...sources) {
    sources.forEach((s) => {
      for (const key in s) {
        if (s[key] && typeof s[key] == "object" && !Array.isArray(s[key])) {
          if (!target[key]) target[key] = {};
          deepMerge(target[key], s[key]);
        } else {
          target[key] = s[key];
        }
      }
    });
    return target;
  }
  function dotNotation(obj, prop, set) {
    if (obj == null || !prop) return void 0;
    return prop.split(/[.[\]]/g).filter((prop2) => prop2.length).reduce((obj2, prop2, i, arr) => {
      if (prop2[0] == '"' || prop2[0] == "'") prop2 = prop2.slice(1, -1);
      if (!(obj2 == null ? void 0 : obj2.hasOwnProperty(prop2))) {
        if (set == void 0) return void 0;
        obj2[prop2] = {};
      }
      if (set !== void 0 && i == arr.length - 1)
        return obj2[prop2] = set;
      return obj2[prop2];
    }, obj);
  }
  function encodeQuery(data) {
    return Object.entries(data).map(
      ([key, value]) => encodeURIComponent(key) + "=" + encodeURIComponent(value)
    ).join("&");
  }
  function flattenObj(obj, parent, result = {}) {
    if (typeof obj === "object" && !Array.isArray(obj)) {
      for (const key of Object.keys(obj)) {
        const propName = parent ? `${parent}.${key}` : key;
        if (typeof obj[key] === "object" && obj[key] != null && !Array.isArray(obj[key])) {
          console.log(propName);
          flattenObj(obj[key], propName, result);
        } else {
          result[propName] = obj[key];
        }
      }
      return result;
    }
  }
  function formData(target) {
    const data = new FormData();
    Object.entries(target).forEach(([key, value]) => data.append(key, value));
    return data;
  }
  function includes(target, values, allowMissing = false) {
    if (target == void 0) return allowMissing;
    if (Array.isArray(values)) return values.findIndex((e, i) => !includes(target[i], values[i], allowMissing)) == -1;
    const type = typeof values;
    if (type != typeof target) return false;
    if (type == "object") {
      return Object.keys(values).find((key) => !includes(target[key], values[key], allowMissing)) == null;
    }
    if (type == "function") return target.toString() == values.toString();
    return target == values;
  }
  function isEqual(a, b) {
    const ta = typeof a, tb = typeof b;
    if (ta != "object" || a == null || (tb != "object" || b == null))
      return ta == "function" && tb == "function" ? a.toString() == b.toString() : a === b;
    const keys = Object.keys(a);
    if (keys.length != Object.keys(b).length) return false;
    return Object.keys(a).every((key) => isEqual(a[key], b[key]));
  }
  function mixin(target, constructors) {
    constructors.forEach((c) => {
      Object.getOwnPropertyNames(c.prototype).forEach((name) => {
        Object.defineProperty(
          target.prototype,
          name,
          Object.getOwnPropertyDescriptor(c.prototype, name) || /* @__PURE__ */ Object.create(null)
        );
      });
    });
  }
  function JSONAttemptParse(json) {
    try {
      return JSON.parse(json);
    } catch {
      return json;
    }
  }
  function JSONSanitize(obj, space) {
    return JSON.stringify(obj, (key, value) => {
      return value;
    }, space);
  }
  function addUnique(array, el) {
    if (array.indexOf(el) === -1) array.push(el);
    return array;
  }
  function arrayDiff(a, b) {
    return makeUnique([
      ...a.filter((v1) => !b.includes((v2) => isEqual(v1, v2))),
      ...b.filter((v1) => !a.includes((v2) => isEqual(v1, v2)))
    ]);
  }
  function caseInsensitiveSort(prop) {
    return function(a, b) {
      const aVal = dotNotation(a, prop);
      const bVal = dotNotation(b, prop);
      if (typeof aVal !== "string" || typeof bVal !== "string") return 1;
      return aVal.toLowerCase().localeCompare(bVal.toLowerCase());
    };
  }
  function findByProp(prop, value) {
    return (v) => isEqual(dotNotation(v, prop), value);
  }
  function flattenArr(arr, result = []) {
    arr.forEach((el) => Array.isArray(el) ? flattenArr(el, result) : result.push(el));
    return result;
  }
  function sortByProp(prop, reverse = false) {
    return function(a, b) {
      const aVal = dotNotation(a, prop);
      const bVal = dotNotation(b, prop);
      if (typeof aVal == "number" && typeof bVal == "number")
        return (reverse ? -1 : 1) * (aVal - bVal);
      if (aVal > bVal) return reverse ? -1 : 1;
      if (aVal < bVal) return reverse ? 1 : -1;
      return 0;
    };
  }
  function makeUnique(arr) {
    for (let i = arr.length - 1; i >= 0; i--) {
      if (arr.slice(0, i).find((n) => isEqual(n, arr[i]))) arr.splice(i, 1);
    }
    return arr;
  }
  function makeArray(value) {
    return Array.isArray(value) ? value : [value];
  }
  class ASet extends Array {
    /** Number of elements in set */
    get size() {
      return this.length;
    }
    /**
     * Array to create set from, duplicate values will be removed
     * @param {T[]} elements Elements which will be added to set
     */
    constructor(elements = []) {
      super();
      if (!!(elements == null ? void 0 : elements["forEach"]))
        elements.forEach((el) => this.add(el));
    }
    /**
     * Add elements to set if unique
     * @param items
     */
    add(...items) {
      items.filter((el) => !this.has(el)).forEach((el) => this.push(el));
      return this;
    }
    /**
     * Delete elements from set
     * @param items Elements that will be deleted
     */
    delete(...items) {
      items.forEach((el) => {
        const index = this.indexOf(el);
        if (index != -1) this.splice(index, 1);
      });
      return this;
    }
    /**
     * Create list of elements this set has which the comparison set does not
     * @param {ASet<T>} set Set to compare against
     * @return {ASet<T>} Different elements
     */
    difference(set) {
      return new ASet(this.filter((el) => !set.has(el)));
    }
    /**
     * Check if set includes element
     * @param {T} el Element to look for
     * @return {boolean} True if element was found, false otherwise
     */
    has(el) {
      return this.indexOf(el) != -1;
    }
    /**
     * Find index number of element, or -1 if it doesn't exist. Matches by equality not reference
     *
     * @param {T} search Element to find
     * @param {number} fromIndex Starting index position
     * @return {number} Element index number or -1 if missing
     */
    indexOf(search, fromIndex) {
      return super.findIndex((el) => isEqual(el, search), fromIndex);
    }
    /**
     * Create list of elements this set has in common with the comparison set
     * @param {ASet<T>} set Set to compare against
     * @return {boolean} Set of common elements
     */
    intersection(set) {
      return new ASet(this.filter((el) => set.has(el)));
    }
    /**
     * Check if this set has no elements in common with the comparison set
     * @param {ASet<T>} set Set to compare against
     * @return {boolean} True if nothing in common, false otherwise
     */
    isDisjointFrom(set) {
      return this.intersection(set).size == 0;
    }
    /**
     * Check if all elements in this set are included in the comparison set
     * @param {ASet<T>} set Set to compare against
     * @return {boolean} True if all elements are included, false otherwise
     */
    isSubsetOf(set) {
      return this.findIndex((el) => !set.has(el)) == -1;
    }
    /**
     * Check if all elements from comparison set are included in this set
     * @param {ASet<T>} set Set to compare against
     * @return {boolean} True if all elements are included, false otherwise
     */
    isSuperset(set) {
      return set.findIndex((el) => !this.has(el)) == -1;
    }
    /**
     * Create list of elements that are only in one set but not both (XOR)
     * @param {ASet<T>} set Set to compare against
     * @return {ASet<T>} New set of unique elements
     */
    symmetricDifference(set) {
      return new ASet([...this.difference(set), ...set.difference(this)]);
    }
    /**
     * Create joined list of elements included in this & the comparison set
     * @param {ASet<T>} set Set join
     * @return {ASet<T>} New set of both previous sets combined
     */
    union(set) {
      return new ASet([...this, ...set]);
    }
  }
  class Cache {
    /**
     * Create new cache
     *
     * @param {keyof T} key Default property to use as primary key
     * @param options
     */
    constructor(key, options = {}) {
      __publicField(this, "store", {});
      /** Whether cache is complete */
      __publicField(this, "complete", false);
      /**
       * Get all cached items
       *
       * @return {T[]} Array of items
       */
      __publicField(this, "values", this.all());
      this.key = key;
      this.options = options;
      if (options.storageKey && !options.storage && typeof Storage !== "undefined")
        options.storage = localStorage;
      if (options.storageKey && options.storage) {
        const stored = options.storage.getItem(options.storageKey);
        if (stored) {
          try {
            Object.assign(this.store, JSON.parse(stored));
          } catch {
          }
        }
      }
      return new Proxy(this, {
        get: (target, prop) => {
          if (prop in target) return target[prop];
          return target.store[prop];
        },
        set: (target, prop, value) => {
          if (prop in target) target[prop] = value;
          else target.store[prop] = value;
          return true;
        }
      });
    }
    getKey(value) {
      if (!this.key) throw new Error("No key defined");
      return value[this.key];
    }
    /**
     * Get all cached items
     *
     * @return {T[]} Array of items
     */
    all() {
      return Object.values(this.store);
    }
    /**
     * Add a new item to the cache. Like set, but finds key automatically
     *
     * @param {T} value Item to add to cache
     * @param {number | undefined} ttl Override default expiry
     * @return {this}
     */
    add(value, ttl = this.ttl) {
      const key = this.getKey(value);
      this.set(key, value, ttl);
      return this;
    }
    /**
     * Add several rows to the cache
     *
     * @param {T[]} rows Several items that will be cached using the default key
     * @param complete Mark cache as complete & reliable, defaults to true
     * @return {this}
     */
    addAll(rows, complete = true) {
      rows.forEach((r) => this.add(r));
      this.complete = complete;
      return this;
    }
    /**
     * Remove all keys from cache
     */
    clear() {
      this.store = {};
    }
    /**
     * Delete an item from the cache
     *
     * @param {K} key Item's primary key
     */
    delete(key) {
      delete this.store[key];
      if (this.options.storageKey && this.options.storage)
        this.options.storage.setItem(this.options.storageKey, JSON.stringify(this.store));
    }
    /**
     * Return cache as an array of key-value pairs
     * @return {[K, T][]} Key-value pairs array
     */
    entries() {
      return Object.entries(this.store);
    }
    /**
     * Get item from the cache
     * @param {K} key Key to lookup
     * @return {T} Cached item
     */
    get(key) {
      return this.store[key];
    }
    /**
     * Get a list of cached keys
     *
     * @return {K[]} Array of keys
     */
    keys() {
      return Object.keys(this.store);
    }
    /**
     * Get map of cached items
     *
     * @return {Record<K, T>}
     */
    map() {
      return structuredClone(this.store);
    }
    /**
     * Add an item to the cache manually specifying the key
     *
     * @param {K} key Key item will be cached under
     * @param {T} value Item to cache
     * @param {number | undefined} ttl Override default expiry in seconds
     * @return {this}
     */
    set(key, value, ttl = this.options.ttl) {
      this.store[key] = value;
      if (this.options.storageKey && this.options.storage)
        this.options.storage.setItem(this.options.storageKey, JSON.stringify(this.store));
      if (ttl) setTimeout(() => {
        this.complete = false;
        this.delete(key);
      }, ttl * 1e3);
      return this;
    }
  }
  const LETTER_LIST = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
  const NUMBER_LIST = "0123456789";
  const SYMBOL_LIST = "~`!@#$%^&*()_-+={[}]|\\:;\"'<,>.?/";
  const CHAR_LIST = LETTER_LIST + LETTER_LIST.toLowerCase() + NUMBER_LIST + SYMBOL_LIST;
  function formatBytes(bytes, decimals = 2) {
    if (bytes === 0) return "0 Bytes";
    const k = 1024;
    const sizes = ["Bytes", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB"];
    const i = Math.floor(Math.log(bytes) / Math.log(k));
    return parseFloat((bytes / Math.pow(k, i)).toFixed(decimals)) + " " + sizes[i];
  }
  function formatPhoneNumber(number) {
    const parts = /(\+?1)?.*?(\d{3}).*?(\d{3}).*?(\d{4})/g.exec(number);
    if (!parts) throw new Error(`Number cannot be parsed: ${number}`);
    return `${parts[1] ? "+1" : ""} (${parts[2]}) ${parts[3]}-${parts[4]}`.trim();
  }
  function insertAt(target, str, index) {
    return `${target.slice(0, index)}${str}${target.slice(index + 1)}`;
  }
  function pad(text, length, char = " ", start = true) {
    if (start) return text.toString().padStart(length, char);
    return text.toString().padEnd(length, char);
  }
  function randomHex(length) {
    return Array(length).fill(null).map(() => Math.round(Math.random() * 15).toString(16)).join("");
  }
  function randomString(length, pool = CHAR_LIST) {
    return Array(length).fill(null).map(() => {
      const n = ~~(Math.random() * pool.length);
      return pool[n];
    }).join("");
  }
  function randomStringBuilder(length, letters = false, numbers = false, symbols = false) {
    if (!letters && !numbers && !symbols) throw new Error("Must enable at least one: letters, numbers, symbols");
    return Array(length).fill(null).map(() => {
      let c;
      do {
        const type = ~~(Math.random() * 3);
        if (letters && type == 0) {
          c = LETTER_LIST[~~(Math.random() * LETTER_LIST.length)];
        } else if (numbers && type == 1) {
          c = NUMBER_LIST[~~(Math.random() * NUMBER_LIST.length)];
        } else if (symbols && type == 2) {
          c = SYMBOL_LIST[~~(Math.random() * SYMBOL_LIST.length)];
        }
      } while (!c);
      return c;
    }).join("");
  }
  function strSplice(str, start, deleteCount, insert = "") {
    const before = str.slice(0, start);
    const after = str.slice(start + deleteCount);
    return before + insert + after;
  }
  function matchAll(value, regex) {
    if (typeof regex === "string") {
      regex = new RegExp(regex, "g");
    }
    if (!regex.global) {
      throw new TypeError("Regular expression must be global.");
    }
    let ret = [];
    let match;
    while ((match = regex.exec(value)) !== null) {
      ret.push(match);
    }
    return ret;
  }
  function parseUrl(url) {
    const processed = new RegExp(
      "(?:(?<protocol>[\\w\\d]+)\\:\\/\\/)?(?:(?<user>.+)\\@)?(?<host>(?<domain>[^:\\/\\?#@\\n]+)(?:\\:(?<port>\\d*))?)(?<path>\\/.*?)?(?:\\?(?<query>.*?))?(?:#(?<fragment>.*?))?$",
      "gm"
    ).exec(url);
    const groups = (processed == null ? void 0 : processed.groups) ?? {};
    const domains = groups.domain.split(".");
    if (groups["port"] != null) groups.port = Number(groups.port);
    if (domains.length > 2) {
      groups.domain = domains.splice(-2, 2).join(".");
      groups.subdomain = domains.join(".");
    }
    if (groups.query) {
      const split = groups.query.split("&"), query = {};
      split.forEach((q) => {
        const [key, val] = q.split("=");
        query[key] = val;
      });
      groups.query = query;
    }
    return groups;
  }
  function md5(d) {
    var r = M(V(Y(X(d), 8 * d.length)));
    return r.toLowerCase();
  }
  function M(d) {
    for (var _, m = "0123456789ABCDEF", f = "", r = 0; r < d.length; r++) _ = d.charCodeAt(r), f += m.charAt(_ >>> 4 & 15) + m.charAt(15 & _);
    return f;
  }
  function X(d) {
    for (var _ = Array(d.length >> 2), m = 0; m < _.length; m++) _[m] = 0;
    for (m = 0; m < 8 * d.length; m += 8) _[m >> 5] |= (255 & d.charCodeAt(m / 8)) << m % 32;
    return _;
  }
  function V(d) {
    for (var _ = "", m = 0; m < 32 * d.length; m += 8) _ += String.fromCharCode(d[m >> 5] >>> m % 32 & 255);
    return _;
  }
  function Y(d, _) {
    d[_ >> 5] |= 128 << _ % 32, d[14 + (_ + 64 >>> 9 << 4)] = _;
    for (var m = 1732584193, f = -271733879, r = -1732584194, i = 271733878, n = 0; n < d.length; n += 16) {
      var h = m, t = f, g = r, e = i;
      f = md5_ii(f = md5_ii(f = md5_ii(f = md5_ii(f = md5_hh(f = md5_hh(f = md5_hh(f = md5_hh(f = md5_gg(f = md5_gg(f = md5_gg(f = md5_gg(f = md5_ff(f = md5_ff(f = md5_ff(f = md5_ff(f, r = md5_ff(r, i = md5_ff(i, m = md5_ff(m, f, r, i, d[n + 0], 7, -680876936), f, r, d[n + 1], 12, -389564586), m, f, d[n + 2], 17, 606105819), i, m, d[n + 3], 22, -1044525330), r = md5_ff(r, i = md5_ff(i, m = md5_ff(m, f, r, i, d[n + 4], 7, -176418897), f, r, d[n + 5], 12, 1200080426), m, f, d[n + 6], 17, -1473231341), i, m, d[n + 7], 22, -45705983), r = md5_ff(r, i = md5_ff(i, m = md5_ff(m, f, r, i, d[n + 8], 7, 1770035416), f, r, d[n + 9], 12, -1958414417), m, f, d[n + 10], 17, -42063), i, m, d[n + 11], 22, -1990404162), r = md5_ff(r, i = md5_ff(i, m = md5_ff(m, f, r, i, d[n + 12], 7, 1804603682), f, r, d[n + 13], 12, -40341101), m, f, d[n + 14], 17, -1502002290), i, m, d[n + 15], 22, 1236535329), r = md5_gg(r, i = md5_gg(i, m = md5_gg(m, f, r, i, d[n + 1], 5, -165796510), f, r, d[n + 6], 9, -1069501632), m, f, d[n + 11], 14, 643717713), i, m, d[n + 0], 20, -373897302), r = md5_gg(r, i = md5_gg(i, m = md5_gg(m, f, r, i, d[n + 5], 5, -701558691), f, r, d[n + 10], 9, 38016083), m, f, d[n + 15], 14, -660478335), i, m, d[n + 4], 20, -405537848), r = md5_gg(r, i = md5_gg(i, m = md5_gg(m, f, r, i, d[n + 9], 5, 568446438), f, r, d[n + 14], 9, -1019803690), m, f, d[n + 3], 14, -187363961), i, m, d[n + 8], 20, 1163531501), r = md5_gg(r, i = md5_gg(i, m = md5_gg(m, f, r, i, d[n + 13], 5, -1444681467), f, r, d[n + 2], 9, -51403784), m, f, d[n + 7], 14, 1735328473), i, m, d[n + 12], 20, -1926607734), r = md5_hh(r, i = md5_hh(i, m = md5_hh(m, f, r, i, d[n + 5], 4, -378558), f, r, d[n + 8], 11, -2022574463), m, f, d[n + 11], 16, 1839030562), i, m, d[n + 14], 23, -35309556), r = md5_hh(r, i = md5_hh(i, m = md5_hh(m, f, r, i, d[n + 1], 4, -1530992060), f, r, d[n + 4], 11, 1272893353), m, f, d[n + 7], 16, -155497632), i, m, d[n + 10], 23, -1094730640), r = md5_hh(r, i = md5_hh(i, m = md5_hh(m, f, r, i, d[n + 13], 4, 681279174), f, r, d[n + 0], 11, -358537222), m, f, d[n + 3], 16, -722521979), i, m, d[n + 6], 23, 76029189), r = md5_hh(r, i = md5_hh(i, m = md5_hh(m, f, r, i, d[n + 9], 4, -640364487), f, r, d[n + 12], 11, -421815835), m, f, d[n + 15], 16, 530742520), i, m, d[n + 2], 23, -995338651), r = md5_ii(r, i = md5_ii(i, m = md5_ii(m, f, r, i, d[n + 0], 6, -198630844), f, r, d[n + 7], 10, 1126891415), m, f, d[n + 14], 15, -1416354905), i, m, d[n + 5], 21, -57434055), r = md5_ii(r, i = md5_ii(i, m = md5_ii(m, f, r, i, d[n + 12], 6, 1700485571), f, r, d[n + 3], 10, -1894986606), m, f, d[n + 10], 15, -1051523), i, m, d[n + 1], 21, -2054922799), r = md5_ii(r, i = md5_ii(i, m = md5_ii(m, f, r, i, d[n + 8], 6, 1873313359), f, r, d[n + 15], 10, -30611744), m, f, d[n + 6], 15, -1560198380), i, m, d[n + 13], 21, 1309151649), r = md5_ii(r, i = md5_ii(i, m = md5_ii(m, f, r, i, d[n + 4], 6, -145523070), f, r, d[n + 11], 10, -1120210379), m, f, d[n + 2], 15, 718787259), i, m, d[n + 9], 21, -343485551), m = safe_add(m, h), f = safe_add(f, t), r = safe_add(r, g), i = safe_add(i, e);
    }
    return Array(m, f, r, i);
  }
  function md5_cmn(d, _, m, f, r, i) {
    return safe_add(bit_rol(safe_add(safe_add(_, d), safe_add(f, i)), r), m);
  }
  function md5_ff(d, _, m, f, r, i, n) {
    return md5_cmn(_ & m | ~_ & f, d, _, r, i, n);
  }
  function md5_gg(d, _, m, f, r, i, n) {
    return md5_cmn(_ & f | m & ~f, d, _, r, i, n);
  }
  function md5_hh(d, _, m, f, r, i, n) {
    return md5_cmn(_ ^ m ^ f, d, _, r, i, n);
  }
  function md5_ii(d, _, m, f, r, i, n) {
    return md5_cmn(m ^ (_ | ~f), d, _, r, i, n);
  }
  function safe_add(d, _) {
    var m = (65535 & d) + (65535 & _);
    return (d >> 16) + (_ >> 16) + (m >> 16) << 16 | 65535 & m;
  }
  function bit_rol(d, _) {
    return d << _ | d >>> 32 - _;
  }
  function validateEmail(email) {
    return /^[a-zA-Z0-9.!#$%&'*+/=?^_`{|}~-]+@[a-zA-Z0-9-]+(?:\.[a-zA-Z0-9-]+)*$/.test(email);
  }
  function fromCsv(csv, hasHeaders = true) {
    const row = csv.split("\n");
    let headers = hasHeaders ? row.splice(0, 1)[0] : null;
    if (headers) headers = headers.match(/(?:[^,"']+|"[^"]*"|'[^']*')+/g);
    return row.map((r) => {
      function parseLine(line) {
        const parts = line.split(","), columns = [];
        let quoted = false;
        for (const p of parts) {
          if (quoted) columns[columns.length - 1] = columns.at(-1) + "," + p;
          else columns.push(p);
          if (/[^"]"$/g.test(p)) {
            quoted = false;
          } else if (/^"[^"]/g.test(p)) {
            quoted = true;
          }
        }
        return columns;
      }
      const props = parseLine(r);
      const h = headers || Array(props.length).fill(null).map((r2, i) => {
        let letter = "";
        const first = i / 26;
        if (first > 1) letter += LETTER_LIST[Math.floor(first - 1)];
        letter += LETTER_LIST[i % 26];
        return letter;
      });
      return h.reduce((acc, h2, i) => {
        dotNotation(acc, h2, props[i]);
        return acc;
      }, {});
    });
  }
  function toCsv(target, flatten = true) {
    const headers = new ASet(target.reduce((acc, row) => [...acc, ...Object.keys(flatten ? flattenObj(row) : row)], []));
    return [
      headers.join(","),
      ...target.map((row) => headers.map((h) => {
        const value = dotNotation(row, h);
        return typeof value == "object" && value != null ? '"' + JSONSanitize(value).replaceAll('"', '""') + '"' : value;
      }).join(","))
    ].join("\n");
  }
  class PromiseProgress extends Promise {
    constructor(executor) {
      super((resolve, reject) => executor(
        (value) => resolve(value),
        (reason) => reject(reason),
        (progress) => this.progress = progress
      ));
      __publicField(this, "listeners", []);
      __publicField(this, "_progress", 0);
    }
    get progress() {
      return this._progress;
    }
    set progress(p) {
      if (p == this._progress) return;
      this._progress = p;
      this.listeners.forEach((l) => l(p));
    }
    static from(promise) {
      if (promise instanceof PromiseProgress) return promise;
      return new PromiseProgress((res, rej) => promise.then((...args) => res(...args)).catch((...args) => rej(...args)));
    }
    from(promise) {
      const newPromise = PromiseProgress.from(promise);
      this.onProgress((p) => newPromise.progress = p);
      return newPromise;
    }
    onProgress(callback) {
      this.listeners.push(callback);
      return this;
    }
    then(res, rej) {
      const resp = super.then(res, rej);
      return this.from(resp);
    }
    catch(rej) {
      return this.from(super.catch(rej));
    }
    finally(res) {
      return this.from(super.finally(res));
    }
  }
  function downloadFile(blob, name) {
    if (!(blob instanceof Blob)) blob = new Blob(makeArray(blob));
    const url = URL.createObjectURL(blob);
    downloadUrl(url, name);
    URL.revokeObjectURL(url);
  }
  function downloadUrl(href, name) {
    const a = document.createElement("a");
    a.href = href;
    a.download = name || href.split("/").pop();
    document.body.appendChild(a);
    a.click();
    document.body.removeChild(a);
  }
  function fileBrowser(options = {}) {
    return new Promise((res) => {
      const input = document.createElement("input");
      input.type = "file";
      input.accept = options.accept || "*";
      input.style.display = "none";
      input.multiple = !!options.multiple;
      input.onblur = input.onchange = async () => {
        res(Array.from(input.files));
        input.remove();
      };
      document.body.appendChild(input);
      input.click();
    });
  }
  function fileText(file) {
    return new Promise((resolve, reject) => {
      const reader = new FileReader();
      reader.onload = () => resolve(reader.result);
      reader.onerror = () => reject(reader.error);
      reader.readAsText(file);
    });
  }
  function timestampFilename(name, date = /* @__PURE__ */ new Date()) {
    if (typeof date == "number" || typeof date == "string") date = new Date(date);
    const timestamp = `${date.getFullYear()}-${(date.getMonth() + 1).toString().padStart(2, "0")}-${date.getDate().toString().padStart(2, "0")}_${date.getHours().toString().padStart(2, "0")}-${date.getMinutes().toString().padStart(2, "0")}-${date.getSeconds().toString().padStart(2, "0")}`;
    return name ? name.replace("{{TIMESTAMP}}", timestamp) : timestamp;
  }
  function uploadWithProgress(options) {
    return new PromiseProgress((res, rej, prog) => {
      const xhr = new XMLHttpRequest();
      const formData2 = new FormData();
      options.files.forEach((f) => formData2.append("file", f));
      xhr.withCredentials = !!options.withCredentials;
      xhr.upload.addEventListener("progress", (event) => event.lengthComputable ? prog(event.loaded / event.total) : null);
      xhr.addEventListener("loadend", () => res(JSONAttemptParse(xhr.responseText)));
      xhr.addEventListener("error", () => rej(JSONAttemptParse(xhr.responseText)));
      xhr.addEventListener("timeout", () => rej({ error: "Request timed out" }));
      xhr.open("POST", options.url);
      Object.entries(options.headers || {}).forEach(([key, value]) => xhr.setRequestHeader(key, value));
      xhr.send(formData2);
    });
  }
  class TypedEmitter {
    constructor() {
      __publicField(this, "listeners", {});
    }
    static emit(event, ...args) {
      (this.listeners["*"] || []).forEach((l) => l(event, ...args));
      (this.listeners[event.toString()] || []).forEach((l) => l(...args));
    }
    static off(event, listener) {
      const e = event.toString();
      this.listeners[e] = (this.listeners[e] || []).filter((l) => l === listener);
    }
    static on(event, listener) {
      var _a;
      const e = event.toString();
      if (!this.listeners[e]) this.listeners[e] = [];
      (_a = this.listeners[e]) == null ? void 0 : _a.push(listener);
      return () => this.off(event, listener);
    }
    static once(event, listener) {
      return new Promise((res) => {
        const unsubscribe = this.on(event, (...args) => {
          res(args.length == 1 ? args[0] : args);
          if (listener) listener(...args);
          unsubscribe();
        });
      });
    }
    emit(event, ...args) {
      (this.listeners["*"] || []).forEach((l) => l(event, ...args));
      (this.listeners[event] || []).forEach((l) => l(...args));
    }
    off(event, listener) {
      this.listeners[event] = (this.listeners[event] || []).filter((l) => l === listener);
    }
    on(event, listener) {
      var _a;
      if (!this.listeners[event]) this.listeners[event] = [];
      (_a = this.listeners[event]) == null ? void 0 : _a.push(listener);
      return () => this.off(event, listener);
    }
    once(event, listener) {
      return new Promise((res) => {
        const unsubscribe = this.on(event, (...args) => {
          res(args.length == 1 ? args[0] : args);
          if (listener) listener(...args);
          unsubscribe();
        });
      });
    }
  }
  __publicField(TypedEmitter, "listeners", {});
  class CustomError extends Error {
    constructor(message, code) {
      super(message);
      __publicField(this, "_code");
      if (code != null) this._code = code;
    }
    get code() {
      return this._code || this.constructor.code;
    }
    set code(c) {
      this._code = c;
    }
    static from(err) {
      const code = Number(err.statusCode) ?? Number(err.code);
      const newErr = new this(err.message || err.toString());
      return Object.assign(newErr, {
        stack: err.stack,
        ...err,
        code: code ?? void 0
      });
    }
    static instanceof(err) {
      return err.constructor.code != void 0;
    }
    toString() {
      return this.message || super.toString();
    }
  }
  __publicField(CustomError, "code", 500);
  class BadRequestError extends CustomError {
    constructor(message = "Bad Request") {
      super(message);
    }
    static instanceof(err) {
      return err.constructor.code == this.code;
    }
  }
  __publicField(BadRequestError, "code", 400);
  class UnauthorizedError extends CustomError {
    constructor(message = "Unauthorized") {
      super(message);
    }
    static instanceof(err) {
      return err.constructor.code == this.code;
    }
  }
  __publicField(UnauthorizedError, "code", 401);
  class PaymentRequiredError extends CustomError {
    constructor(message = "Payment Required") {
      super(message);
    }
    static instanceof(err) {
      return err.constructor.code == this.code;
    }
  }
  __publicField(PaymentRequiredError, "code", 402);
  class ForbiddenError extends CustomError {
    constructor(message = "Forbidden") {
      super(message);
    }
    static instanceof(err) {
      return err.constructor.code == this.code;
    }
  }
  __publicField(ForbiddenError, "code", 403);
  class NotFoundError extends CustomError {
    constructor(message = "Not Found") {
      super(message);
    }
    static instanceof(err) {
      return err.constructor.code == this.code;
    }
  }
  __publicField(NotFoundError, "code", 404);
  class MethodNotAllowedError extends CustomError {
    constructor(message = "Method Not Allowed") {
      super(message);
    }
    static instanceof(err) {
      return err.constructor.code == this.code;
    }
  }
  __publicField(MethodNotAllowedError, "code", 405);
  class NotAcceptableError extends CustomError {
    constructor(message = "Not Acceptable") {
      super(message);
    }
    static instanceof(err) {
      return err.constructor.code == this.code;
    }
  }
  __publicField(NotAcceptableError, "code", 406);
  class InternalServerError extends CustomError {
    constructor(message = "Internal Server Error") {
      super(message);
    }
    static instanceof(err) {
      return err.constructor.code == this.code;
    }
  }
  __publicField(InternalServerError, "code", 500);
  class NotImplementedError extends CustomError {
    constructor(message = "Not Implemented") {
      super(message);
    }
    static instanceof(err) {
      return err.constructor.code == this.code;
    }
  }
  __publicField(NotImplementedError, "code", 501);
  class BadGatewayError extends CustomError {
    constructor(message = "Bad Gateway") {
      super(message);
    }
    static instanceof(err) {
      return err.constructor.code == this.code;
    }
  }
  __publicField(BadGatewayError, "code", 502);
  class ServiceUnavailableError extends CustomError {
    constructor(message = "Service Unavailable") {
      super(message);
    }
    static instanceof(err) {
      return err.constructor.code == this.code;
    }
  }
  __publicField(ServiceUnavailableError, "code", 503);
  class GatewayTimeoutError extends CustomError {
    constructor(message = "Gateway Timeout") {
      super(message);
    }
    static instanceof(err) {
      return err.constructor.code == this.code;
    }
  }
  __publicField(GatewayTimeoutError, "code", 504);
  function errorFromCode(code, message) {
    switch (code) {
      case 400:
        return new BadRequestError(message);
      case 401:
        return new UnauthorizedError(message);
      case 402:
        return new PaymentRequiredError(message);
      case 403:
        return new ForbiddenError(message);
      case 404:
        return new NotFoundError(message);
      case 405:
        return new MethodNotAllowedError(message);
      case 406:
        return new NotAcceptableError(message);
      case 500:
        return new InternalServerError(message);
      case 501:
        return new NotImplementedError(message);
      case 502:
        return new BadGatewayError(message);
      case 503:
        return new ServiceUnavailableError(message);
      case 504:
        return new GatewayTimeoutError(message);
      default:
        return new CustomError(message, code);
    }
  }
  const _Http = class _Http {
    constructor(defaults = {}) {
      __publicField(this, "interceptors", {});
      __publicField(this, "headers", {});
      __publicField(this, "url");
      this.url = defaults.url ?? null;
      this.headers = defaults.headers || {};
      if (defaults.interceptors) {
        defaults.interceptors.forEach((i) => _Http.addInterceptor(i));
      }
    }
    static addInterceptor(fn) {
      const key = Object.keys(_Http.interceptors).length.toString();
      _Http.interceptors[key] = fn;
      return () => {
        _Http.interceptors[key] = null;
      };
    }
    addInterceptor(fn) {
      const key = Object.keys(this.interceptors).length.toString();
      this.interceptors[key] = fn;
      return () => {
        this.interceptors[key] = null;
      };
    }
    request(opts = {}) {
      var _a;
      if (!this.url && !opts.url) throw new Error("URL needs to be set");
      let url = (((_a = opts.url) == null ? void 0 : _a.startsWith("http")) ? opts.url : (this.url || "") + (opts.url || "")).replace(/([^:]\/)\/+/g, "$1");
      if (opts.fragment) url.includes("#") ? url.replace(/#.*(\?|\n)/g, (match, arg1) => `#${opts.fragment}${arg1}`) : url += "#" + opts.fragment;
      if (opts.query) {
        const q = Array.isArray(opts.query) ? opts.query : Object.keys(opts.query).map((k) => ({ key: k, value: opts.query[k] }));
        url += (url.includes("?") ? "&" : "?") + q.map((q2) => `${q2.key}=${q2.value}`).join("&");
      }
      const headers = clean({
        "Content-Type": !opts.body ? void 0 : opts.body instanceof FormData ? "multipart/form-data" : "application/json",
        ..._Http.headers,
        ...this.headers,
        ...opts.headers
      });
      if (typeof opts.body == "object" && opts.body != null && headers["Content-Type"] == "application/json")
        opts.body = JSON.stringify(opts.body);
      return new PromiseProgress((res, rej, prog) => {
        try {
          fetch(url, {
            headers,
            method: opts.method || (opts.body ? "POST" : "GET"),
            body: opts.body
          }).then(async (resp) => {
            var _a2, _b;
            for (let fn of [...Object.values(_Http.interceptors), ...Object.values(this.interceptors)]) {
              await new Promise((res2) => fn(resp, () => res2()));
            }
            const contentLength = resp.headers.get("Content-Length");
            const total = contentLength ? parseInt(contentLength, 10) : 0;
            let loaded = 0;
            const reader = (_a2 = resp.body) == null ? void 0 : _a2.getReader();
            const stream = new ReadableStream({
              start(controller) {
                function push() {
                  reader == null ? void 0 : reader.read().then((event) => {
                    if (event.done) return controller.close();
                    loaded += event.value.byteLength;
                    prog(loaded / total);
                    controller.enqueue(event.value);
                    push();
                  }).catch((error) => controller.error(error));
                }
                push();
              }
            });
            resp.data = new Response(stream);
            if (opts.decode == null || opts.decode) {
              const content = (_b = resp.headers.get("Content-Type")) == null ? void 0 : _b.toLowerCase();
              if (content == null ? void 0 : content.includes("form")) resp.data = await resp.data.formData();
              else if (content == null ? void 0 : content.includes("json")) resp.data = await resp.data.json();
              else if (content == null ? void 0 : content.includes("text")) resp.data = await resp.data.text();
              else if (content == null ? void 0 : content.includes("application")) resp.data = await resp.data.blob();
            }
            if (resp.ok) res(resp);
            else rej(resp);
          }).catch((err) => rej(err));
        } catch (err) {
          rej(err);
        }
      });
    }
  };
  __publicField(_Http, "interceptors", {});
  __publicField(_Http, "headers", {});
  let Http = _Http;
  function jwtDecode(token) {
    const base64 = token.split(".")[1].replace(/-/g, "+").replace(/_/g, "/");
    return JSONAttemptParse(decodeURIComponent(atob(base64).split("").map(function(c) {
      return "%" + ("00" + c.charCodeAt(0).toString(16)).slice(-2);
    }).join("")));
  }
  const CliEffects = {
    CLEAR: "\x1B[0m",
    BRIGHT: "\x1B[1m",
    DIM: "\x1B[2m",
    UNDERSCORE: "\x1B[4m",
    BLINK: "\x1B[5m",
    REVERSE: "\x1B[7m",
    HIDDEN: "\x1B[8m"
  };
  const CliForeground = {
    BLACK: "\x1B[30m",
    RED: "\x1B[31m",
    GREEN: "\x1B[32m",
    YELLOW: "\x1B[33m",
    BLUE: "\x1B[34m",
    MAGENTA: "\x1B[35m",
    CYAN: "\x1B[36m",
    LIGHT_GREY: "\x1B[37m",
    GREY: "\x1B[90m",
    LIGHT_RED: "\x1B[91m",
    LIGHT_GREEN: "\x1B[92m",
    LIGHT_YELLOW: "\x1B[93m",
    LIGHT_BLUE: "\x1B[94m",
    LIGHT_MAGENTA: "\x1B[95m",
    LIGHT_CYAN: "\x1B[96m",
    WHITE: "\x1B[97m"
  };
  const CliBackground = {
    BLACK: "\x1B[40m",
    RED: "\x1B[41m",
    GREEN: "\x1B[42m",
    YELLOW: "\x1B[43m",
    BLUE: "\x1B[44m",
    MAGENTA: "\x1B[45m",
    CYAN: "\x1B[46m",
    WHITE: "\x1B[47m",
    GREY: "\x1B[100m"
  };
  var LOG_LEVEL = /* @__PURE__ */ ((LOG_LEVEL2) => {
    LOG_LEVEL2[LOG_LEVEL2["ERROR"] = 0] = "ERROR";
    LOG_LEVEL2[LOG_LEVEL2["WARN"] = 1] = "WARN";
    LOG_LEVEL2[LOG_LEVEL2["INFO"] = 2] = "INFO";
    LOG_LEVEL2[LOG_LEVEL2["LOG"] = 3] = "LOG";
    LOG_LEVEL2[LOG_LEVEL2["DEBUG"] = 4] = "DEBUG";
    return LOG_LEVEL2;
  })(LOG_LEVEL || {});
  const _Logger = class _Logger extends TypedEmitter {
    constructor(namespace) {
      super();
      this.namespace = namespace;
    }
    format(...text) {
      const now = /* @__PURE__ */ new Date();
      const timestamp = `${now.getFullYear()}-${now.getMonth() + 1}-${now.getDate()} ${now.getHours().toString().padStart(2, "0")}:${now.getMinutes().toString().padStart(2, "0")}:${now.getSeconds().toString().padStart(2, "0")}.${now.getMilliseconds().toString().padEnd(3, "0")}`;
      return `${timestamp}${this.namespace ? ` [${this.namespace}]` : ""} ${text.map((t) => typeof t == "string" ? t : JSONSanitize(t, 2)).join(" ")}`;
    }
    debug(...args) {
      if (_Logger.LOG_LEVEL < 4) return;
      const str = this.format(...args);
      _Logger.emit(4, str);
      console.debug(CliForeground.LIGHT_GREY + str + CliEffects.CLEAR);
    }
    log(...args) {
      if (_Logger.LOG_LEVEL < 3) return;
      const str = this.format(...args);
      _Logger.emit(3, str);
      console.log(CliEffects.CLEAR + str);
    }
    info(...args) {
      if (_Logger.LOG_LEVEL < 2) return;
      const str = this.format(...args);
      _Logger.emit(2, str);
      console.info(CliForeground.BLUE + str + CliEffects.CLEAR);
    }
    warn(...args) {
      if (_Logger.LOG_LEVEL < 1) return;
      const str = this.format(...args);
      _Logger.emit(1, str);
      console.warn(CliForeground.YELLOW + str + CliEffects.CLEAR);
    }
    error(...args) {
      if (_Logger.LOG_LEVEL < 0) return;
      const str = this.format(...args);
      _Logger.emit(0, str);
      console.error(CliForeground.RED + str + CliEffects.CLEAR);
    }
  };
  __publicField(_Logger, "LOG_LEVEL", 4);
  let Logger = _Logger;
  function dec2Frac(num) {
    const gcd = (a, b) => {
      if (b < 1e-7) return a;
      return gcd(b, ~~(a % b));
    };
    const len = num.toString().length - 2;
    let denominator = Math.pow(10, len);
    let numerator = num * denominator;
    const divisor = gcd(numerator, denominator);
    numerator = ~~(numerator / divisor);
    denominator = ~~(denominator / divisor);
    const remainder = ~~(numerator / denominator);
    numerator -= remainder * denominator;
    return `${remainder ? remainder + " " : ""}${~~numerator}/${~~denominator}`;
  }
  function fracToDec(frac) {
    let split = frac.split(" ");
    const whole = split.length == 2 ? Number(split[0]) : 0;
    split = split.pop().split("/");
    return whole + Number(split[0]) / Number(split[1]);
  }
  function gravatar(email, def = "mp") {
    if (!email) return "";
    return `https://www.gravatar.com/avatar/${md5(email)}?d=${def}`;
  }
  function escapeRegex(value) {
    return value.replace(/[.*+?^${}()|\[\]\\]/g, "\\$&");
  }
  function PE(str, ...args) {
    const combined = [];
    for (let i = 0; i < str.length || i < args.length; i++) {
      if (str[i]) combined.push(str[i]);
      if (args[i]) combined.push(args[i]);
    }
    return new PathEvent(combined.join(""));
  }
  function PES(str, ...args) {
    let combined = [];
    for (let i = 0; i < str.length || i < args.length; i++) {
      if (str[i]) combined.push(str[i]);
      if (args[i]) combined.push(args[i]);
    }
    const [paths, methods] = combined.join("").split(":");
    return PathEvent.toString(paths, methods == null ? void 0 : methods.split(""));
  }
  class PathError extends Error {
  }
  class PathEvent {
    constructor(Event) {
      /** First directory in path */
      __publicField(this, "module");
      /** Entire path, including the module & name */
      __publicField(this, "fullPath");
      /** Path including the name, excluding the module */
      __publicField(this, "path");
      /** Last sagment of path */
      __publicField(this, "name");
      /** List of methods */
      __publicField(this, "methods");
      var _a;
      if (typeof Event == "object") return Object.assign(this, Event);
      let [p, scope, method] = Event.replaceAll(/\/{2,}/g, "/").split(":");
      if (!method) method = scope || "*";
      if (p == "*" || !p && method == "*") {
        p = "";
        method = "*";
      }
      let temp = p.split("/").filter((p2) => !!p2);
      this.module = ((_a = temp.splice(0, 1)[0]) == null ? void 0 : _a.toLowerCase()) || "";
      this.fullPath = p;
      this.path = temp.join("/");
      this.name = temp.pop() || "";
      this.methods = new ASet(method.split(""));
    }
    /** All/Wildcard specified */
    get all() {
      return this.methods.has("*");
    }
    set all(v) {
      v ? new ASet(["*"]) : this.methods.delete("*");
    }
    /** None specified */
    get none() {
      return this.methods.has("n");
    }
    set none(v) {
      v ? this.methods = new ASet(["n"]) : this.methods.delete("n");
    }
    /** Create method specified */
    get create() {
      return !this.methods.has("n") && (this.methods.has("*") || this.methods.has("c"));
    }
    set create(v) {
      v ? this.methods.delete("n").add("c") : this.methods.delete("c");
    }
    /** Read method specified */
    get read() {
      return !this.methods.has("n") && (this.methods.has("*") || this.methods.has("r"));
    }
    set read(v) {
      v ? this.methods.delete("n").add("r") : this.methods.delete("r");
    }
    /** Update method specified */
    get update() {
      return !this.methods.has("n") && (this.methods.has("*") || this.methods.has("u"));
    }
    set update(v) {
      v ? this.methods.delete("n").add("u") : this.methods.delete("u");
    }
    /** Delete method specified */
    get delete() {
      return !this.methods.has("n") && (this.methods.has("*") || this.methods.has("d"));
    }
    set delete(v) {
      v ? this.methods.delete("n").add("d") : this.methods.delete("d");
    }
    /**
     * Combine multiple  events into one parsed object. Longest path takes precedent, but all subsequent methods are
     * combined until a "none" is reached
     *
     * @param {string | PathEvent} paths Events as strings or pre-parsed
     * @return {PathEvent} Final combined permission
     */
    static combine(...paths) {
      let hitNone = false;
      const combined = paths.map((p) => new PathEvent(p)).toSorted((p1, p2) => {
        const l1 = p1.fullPath.length, l2 = p2.fullPath.length;
        return l1 < l2 ? 1 : l1 > l2 ? -1 : 0;
      }).reduce((acc, p) => {
        if (p.none) hitNone = true;
        if (!acc) return p;
        if (hitNone) return acc;
        acc.methods = [...acc.methods, ...p.methods];
        return acc;
      }, null);
      combined.methods = new ASet(combined.methods);
      return combined;
    }
    /**
     * Filter a set of paths based on the target
     *
     * @param {string | PathEvent | (string | PathEvent)[]} target Array of events that will filtered
     * @param filter {...PathEvent} Must container one of
     * @return {boolean} Whether there is any overlap
     */
    static filter(target, ...filter) {
      const parsedTarget = makeArray(target).map((pe) => new PathEvent(pe));
      const parsedFind = makeArray(filter).map((pe) => new PathEvent(pe));
      return parsedTarget.filter((t) => {
        if (!t.fullPath && t.all) return true;
        return !!parsedFind.find((f) => t.fullPath.startsWith(f.fullPath) && (f.all || t.all || t.methods.intersection(f.methods).length));
      });
    }
    /**
     * Squash 2 sets of paths & return true if any overlap is found
     *
     * @param {string | PathEvent | (string | PathEvent)[]} target Array of Events as strings or pre-parsed
     * @param has Target must have at least one of these path
     * @return {boolean} Whether there is any overlap
     */
    static has(target, ...has) {
      const parsedRequired = makeArray(has).map((pe) => new PathEvent(pe));
      const parsedTarget = makeArray(target).map((pe) => new PathEvent(pe));
      return !!parsedRequired.find((r) => {
        if (!r.fullPath && r.all) return true;
        const filtered = parsedTarget.filter((p) => r.fullPath.startsWith(p.fullPath));
        if (!filtered.length) return false;
        const combined = PathEvent.combine(...filtered);
        return !combined.none && (combined.all || r.all) || combined.methods.intersection(r.methods).length;
      });
    }
    /**
     * Squash 2 sets of paths & return true if the target has all paths
     *
     * @param {string | PathEvent | (string | PathEvent)[]} target Array of Events as strings or pre-parsed
     * @param has Target must have all these paths
     * @return {boolean} Whether there is any overlap
     */
    static hasAll(target, ...has) {
      return has.filter((h) => PathEvent.has(target, h)).length == has.length;
    }
    /**
     * Same as `has` but raises an error if there is no overlap
     *
     * @param {string | string[]} target Array of Events as strings or pre-parsed
     * @param has Target must have at least one of these path
     */
    static hasFatal(target, ...has) {
      if (!PathEvent.has(target, ...has)) throw new PathError(`Requires one of: ${makeArray(has).join(", ")}`);
    }
    /**
     * Same as `hasAll` but raises an error if the target is missing any paths
     *
     * @param {string | string[]} target Array of Events as strings or pre-parsed
     * @param has Target must have all these paths
     */
    static hasAllFatal(target, ...has) {
      if (!PathEvent.hasAll(target, ...has)) throw new PathError(`Requires all: ${makeArray(has).join(", ")}`);
    }
    /**
     * Create  event string from its components
     *
     * @param {string | string[]} path Event path
     * @param {Method} methods Event method
     * @return {string} String representation of Event
     */
    static toString(path, methods) {
      let p = makeArray(path).filter((p2) => p2 != null).join("/");
      p = p == null ? void 0 : p.trim().replaceAll(/\/{2,}/g, "/").replaceAll(/(^\/|\/$)/g, "");
      if (methods == null ? void 0 : methods.length) p += `:${makeArray(methods).map((m) => m.toLowerCase()).join("")}`;
      return p;
    }
    /**
     * Filter a set of paths based on this event
     *
     * @param {string | PathEvent | (string | PathEvent)[]} target Array of events that will filtered
     * @return {boolean} Whether there is any overlap
     */
    filter(target) {
      return PathEvent.filter(target, this);
    }
    /**
     * Create  event string from its components
     *
     * @return {string} String representation of Event
     */
    toString() {
      return PathEvent.toString(this.fullPath, this.methods);
    }
  }
  class PathEventEmitter {
    constructor() {
      __publicField(this, "listeners", []);
    }
    emit(event, ...args) {
      const parsed = new PathEvent(event);
      this.listeners.filter((l) => PathEvent.has(l[0], event)).forEach(async (l) => l[1](parsed, ...args));
    }
    off(listener) {
      this.listeners = this.listeners.filter((l) => l[1] != listener);
    }
    on(event, listener) {
      makeArray(event).forEach((e) => this.listeners.push([new PathEvent(e), listener]));
      return () => this.off(listener);
    }
    once(event, listener) {
      return new Promise((res) => {
        const unsubscribe = this.on(event, (event2, ...args) => {
          res(args.length < 2 ? args[0] : args);
          if (listener) listener(event2, ...args);
          unsubscribe();
        });
      });
    }
    relayEvents(emitter) {
      emitter.on("*", (event, ...args) => this.emit(event, ...args));
    }
  }
  function adjustedInterval(cb, ms) {
    let cancel = false, timeout = null;
    const p = async () => {
      if (cancel) return;
      const start = (/* @__PURE__ */ new Date()).getTime();
      await cb();
      const end = (/* @__PURE__ */ new Date()).getTime();
      timeout = setTimeout(() => p(), ms - (end - start) || 1);
    };
    p();
    return () => {
      cancel = true;
      if (timeout) clearTimeout(timeout);
    };
  }
  function formatDate(date, format = "YYYY-MM-DD H:mm ") {
    if (typeof date == "number" || typeof date == "string") date = new Date(date);
    function day(num) {
      switch (num) {
        case 0:
          return "Sunday";
        case 1:
          return "Monday";
        case 2:
          return "Tuesday";
        case 3:
          return "Wednesday";
        case 4:
          return "Thursday";
        case 5:
          return "Friday";
        case 6:
          return "Saturday";
        default:
          return "Unknown";
      }
    }
    function doy(date2) {
      const start = /* @__PURE__ */ new Date(`${date2.getFullYear()}-01-01 0:00:00`);
      return Math.ceil((date2.getTime() - start.getTime()) / (1e3 * 60 * 60 * 24));
    }
    function month(num) {
      switch (num) {
        case 0:
          return "January";
        case 1:
          return "February";
        case 2:
          return "March";
        case 3:
          return "April";
        case 4:
          return "May";
        case 5:
          return "June";
        case 6:
          return "July";
        case 7:
          return "August";
        case 8:
          return "September";
        case 9:
          return "October";
        case 10:
          return "November";
        case 11:
          return "December";
        default:
          return "Unknown";
      }
    }
    function suffix(num) {
      let n = num.toString();
      switch (n.at(-1)) {
        case "1":
          return num + "st";
        case "2":
          return num + "nd";
        case "3":
          return num + "rd";
        default:
          return num + "th";
      }
    }
    function tzOffset(offset) {
      const hours = ~~(offset / 60);
      const minutes = offset % 60;
      return (offset > 0 ? "-" : "") + `${hours}:${minutes.toString().padStart(2, "0")}`;
    }
    function timezone(date2) {
      const formatter = new Intl.DateTimeFormat("en-US", { timeZoneName: "short" });
      const formattedDate = formatter.format(date2);
      return formattedDate.split(" ").pop() || "";
    }
    return format.replaceAll("YYYY", date.getFullYear().toString()).replaceAll("YY", date.getFullYear().toString().slice(2)).replaceAll("MMMM", month(date.getMonth())).replaceAll("MMM", month(date.getMonth()).slice(0, 2)).replaceAll("MM", (date.getMonth() + 1).toString().padStart(2, "0")).replaceAll("M", (date.getMonth() + 1).toString()).replaceAll("DDD", doy(date).toString()).replaceAll("DD", date.getDate().toString().padStart(2, "0")).replaceAll("Do", suffix(date.getDate())).replaceAll("D", date.getDate().toString()).replaceAll("dddd", day(date.getDay())).replaceAll("ddd", day(date.getDay()).slice(0, 2)).replaceAll("dd", date.getDate().toString().padStart(2, "0")).replaceAll("d", date.getDay().toString()).replaceAll("HH", date.getHours().toString().padStart(2, "0")).replaceAll("H", date.getHours().toString()).replaceAll("hh", (date.getHours() > 12 ? date.getHours() - 12 : date.getHours()).toString().padStart(2, "0")).replaceAll("h", (date.getHours() > 12 ? date.getHours() - 12 : date.getHours()).toString()).replaceAll("mm", date.getMinutes().toString().padStart(2, "0")).replaceAll("m", date.getMinutes().toString()).replaceAll("ss", date.getSeconds().toString().padStart(2, "0")).replaceAll("s", date.getSeconds().toString()).replaceAll("SSS", date.getMilliseconds().toString()).replaceAll("SS", date.getMilliseconds().toString().slice(0, 1)).replaceAll("S", date.getMilliseconds().toString()[0]).replaceAll("A", date.getHours() >= 12 ? "PM" : "AM").replaceAll("a", date.getHours() >= 12 ? "pm" : "am").replaceAll("ZZ", tzOffset(date.getTimezoneOffset()).replace(":", "")).replaceAll("Z", tzOffset(date.getTimezoneOffset())).replaceAll("z", timezone(date));
  }
  function sleep(ms) {
    return new Promise((res) => setTimeout(res, ms));
  }
  async function sleepWhile(fn, checkInterval = 100) {
    while (await fn()) await sleep(checkInterval);
  }
  function timeUntil(date) {
    return (date instanceof Date ? date.getTime() : date) - (/* @__PURE__ */ new Date()).getTime();
  }
  function typeKeys() {
    return Object.keys({});
  }
  exports2.ASet = ASet;
  exports2.BadGatewayError = BadGatewayError;
  exports2.BadRequestError = BadRequestError;
  exports2.CHAR_LIST = CHAR_LIST;
  exports2.Cache = Cache;
  exports2.CliBackground = CliBackground;
  exports2.CliEffects = CliEffects;
  exports2.CliForeground = CliForeground;
  exports2.CustomError = CustomError;
  exports2.ForbiddenError = ForbiddenError;
  exports2.GatewayTimeoutError = GatewayTimeoutError;
  exports2.Http = Http;
  exports2.InternalServerError = InternalServerError;
  exports2.JSONAttemptParse = JSONAttemptParse;
  exports2.JSONSanitize = JSONSanitize;
  exports2.LETTER_LIST = LETTER_LIST;
  exports2.LOG_LEVEL = LOG_LEVEL;
  exports2.Logger = Logger;
  exports2.MethodNotAllowedError = MethodNotAllowedError;
  exports2.NUMBER_LIST = NUMBER_LIST;
  exports2.NotAcceptableError = NotAcceptableError;
  exports2.NotFoundError = NotFoundError;
  exports2.NotImplementedError = NotImplementedError;
  exports2.PE = PE;
  exports2.PES = PES;
  exports2.PathError = PathError;
  exports2.PathEvent = PathEvent;
  exports2.PathEventEmitter = PathEventEmitter;
  exports2.PaymentRequiredError = PaymentRequiredError;
  exports2.PromiseProgress = PromiseProgress;
  exports2.SYMBOL_LIST = SYMBOL_LIST;
  exports2.ServiceUnavailableError = ServiceUnavailableError;
  exports2.TypedEmitter = TypedEmitter;
  exports2.UnauthorizedError = UnauthorizedError;
  exports2.addUnique = addUnique;
  exports2.adjustedInterval = adjustedInterval;
  exports2.arrayDiff = arrayDiff;
  exports2.caseInsensitiveSort = caseInsensitiveSort;
  exports2.clean = clean;
  exports2.dec2Frac = dec2Frac;
  exports2.deepCopy = deepCopy;
  exports2.deepMerge = deepMerge;
  exports2.dotNotation = dotNotation;
  exports2.downloadFile = downloadFile;
  exports2.downloadUrl = downloadUrl;
  exports2.encodeQuery = encodeQuery;
  exports2.errorFromCode = errorFromCode;
  exports2.escapeRegex = escapeRegex;
  exports2.fileBrowser = fileBrowser;
  exports2.fileText = fileText;
  exports2.findByProp = findByProp;
  exports2.flattenArr = flattenArr;
  exports2.flattenObj = flattenObj;
  exports2.formData = formData;
  exports2.formatBytes = formatBytes;
  exports2.formatDate = formatDate;
  exports2.formatPhoneNumber = formatPhoneNumber;
  exports2.fracToDec = fracToDec;
  exports2.fromCsv = fromCsv;
  exports2.gravatar = gravatar;
  exports2.includes = includes;
  exports2.insertAt = insertAt;
  exports2.isEqual = isEqual;
  exports2.jwtDecode = jwtDecode;
  exports2.makeArray = makeArray;
  exports2.makeUnique = makeUnique;
  exports2.matchAll = matchAll;
  exports2.md5 = md5;
  exports2.mixin = mixin;
  exports2.pad = pad;
  exports2.parseUrl = parseUrl;
  exports2.randomHex = randomHex;
  exports2.randomString = randomString;
  exports2.randomStringBuilder = randomStringBuilder;
  exports2.sleep = sleep;
  exports2.sleepWhile = sleepWhile;
  exports2.sortByProp = sortByProp;
  exports2.strSplice = strSplice;
  exports2.timeUntil = timeUntil;
  exports2.timestampFilename = timestampFilename;
  exports2.toCsv = toCsv;
  exports2.typeKeys = typeKeys;
  exports2.uploadWithProgress = uploadWithProgress;
  exports2.validateEmail = validateEmail;
  Object.defineProperty(exports2, Symbol.toStringTag, { value: "Module" });
});
//# sourceMappingURL=index.cjs.map
